from setuptools import find_packages, setup

setup(
    name="meegnet",
    packages=find_packages(),
    version="0.2.4",
    description="A set of MEG-optimized ANNs and visualization tools.",
    author="Arthur Dehgan",
    license="MIT",
)
